import numpy as np
import pandas as pd
import torch
from functools import partial
from torch import tensor
from torch.utils.data import DataLoader
from .optimizer import adam_opt
from .utils import ALL_CALLBACKS, param_getter, listify, CancelTrainException, CancelEpochException, CancelBatchException, CancelPredsException
from .callbacks import TrainEvalCallback, ProgressCallback, Plotter, SendToDeviceCallback, AvgStatsCallback, SaveModelCallback, GetPredictionsCallback

class Learner():
    """
    Main train-eval class that packages together a model, the data,
    the loss function and the optimizer.
    In addition, it keeps track of a learning rate and a splitter function that
    can split the model layers into mutiple groups so that they can be trained
    with different learning rates.
    Default splitter does nothing, but this is where splitter functions for
    discriminative learning rates should be passed.
    Finally, the learner is passed a list of callbacks and can be called (like
    a function) with attrs inherited from the callbacks.
    Callbacks added by default: TrainEvalCallback, ProgressCallback, Plotter,
    SendToDeviceCallback (send to cuda if available), AvgStatsCallback for loss_func
    and metrics.
    """
    def __init__(self, model, data, loss_func, opt_func=adam_opt(), lr=1e-2,
                 splitter=param_getter, metrics=None, callbacks=None,
                 callback_funcs=None, device=None, reset_opt=False,
                 model_name=None, path_str=None):
        self.model = model
        self.data = data
        self.loss_func = loss_func
        self.opt_func = opt_func
        self.lr = lr
        self.splitter = splitter
        self.metrics = metrics
        if device == None:
            # by default, we select the first GPU or the CPU if there are no GPUs
            self.device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
        else:
            # use the device argument of the Learner to select a specific GPU
            # for example, 'cuda:0' will select the first GPU, 'cuda:1' the second etc.
            self.device = torch.device(device)
        self.reset_opt = reset_opt # whether to reset the optimizer or not; resets after LRFinder
        # this is where we will save the metric_dicts contributed by AvgStatsCallback or GetPredictionsCallback
        self.train_metrics_dict = None
        self.valid_metrics_dict = None
        self.pred_metrics_dict = None
        self.pred_avg_stats = None
        self.pred_df = pd.DataFrame(columns=[#'X',
                                             'Y', 'Y_hat'])
        self.in_train = False
        self.logger = print
        self.opt = None
        self.ALL_CALLBACKS = ALL_CALLBACKS # inventory of all possible callback functions
        self.callbacks = [] # the list of callbacks that are actually run

        # self.callback_list stores default callbacks to be added to self.callbacks below
        self.callback_list = [ProgressCallback(), Plotter(), SendToDeviceCallback()]
        if model_name and isinstance(model_name, str):
            self.callback_list.append(SaveModelCallback(model_name=model_name, path_str=path_str))
        # add the callbacks called with the Learner instance to the self.callback_list
        self.callback_list.extend(listify(callbacks))
        # add the callback_funcs called with the Learner instance to the self.callback_list
        self.callback_list.extend(callback_func() for callback_func in listify(callback_funcs))

        # add the self.callback_list to self.callbacks
        self.add_callbacks(self.callback_list)

    def add_callbacks(self, callbacks):
        for callback in listify(callbacks):
            self.add_callback(callback)

    def add_callback(self, callback):
        "grab callback and set it as an attr under its name"
        callback.set_learner(self)
        setattr(self, callback.name, callback)
        self.callbacks.append(callback)

    def remove_callbacks(self, callbacks):
        for callback in listify(callbacks):
            self.callbacks.remove(callback)

    def one_batch(self, i, xb, yb):
        try:
            self.iter = i
            self.xb, self.yb = xb, yb;                      self('begin_batch')
            self.pred = self.model(self.xb);                self('after_pred')
            self.loss = self.loss_func(self.pred, self.yb); self('after_loss')
            if not self.in_train:
                return
            self.loss.backward();                           self('after_backward')
            self.opt.step();                                self('after_step')
            self.opt.zero_grad()
        except CancelBatchException:                        self('after_cancel_batch')
        finally:                                            self('after_batch')

    def all_batches(self):
        self.iters = len(self.dl)
        try:
            for i, (xb, yb) in enumerate(self.dl):
                self.one_batch(i, xb, yb)
        except CancelEpochException: self('after_cancel_epoch')

    def do_begin_fit(self, epochs):
        self.epochs = epochs
        self.loss = tensor(0.)
        self('begin_fit')

    def do_begin_epoch(self, epoch):
        self.epoch = epoch
        self.dl = self.data.train_dl
        return self('begin_epoch')

    def fit(self, epochs, callbacks=None):
        # add TrainEvalCallback and AvgStatsCallback to the fit_callbacks
        fit_callbacks = [TrainEvalCallback(),
                         AvgStatsCallback(listify(self.metrics))]
        # add the callbacks passed to fit() to callbacks
        fit_callbacks.extend(listify(callbacks))
        self.add_callbacks(fit_callbacks)
        # create optimizer on fit(), optionally replacing existing optimizer (e.g., after LRFinder)
        if self.reset_opt or not self.opt:
            self.opt = self.opt_func(self.splitter(self.model), lr=self.lr)
            self.reset_opt = False
        try:
            self.do_begin_fit(epochs)
            for epoch in range(epochs):
                self.do_begin_epoch(epoch)
                if not self('begin_epoch'): self.all_batches()

                with torch.no_grad():
                    self.dl = self.data.valid_dl
                    if not self('begin_validate'): self.all_batches()
                self('after_epoch')
        except CancelTrainException:
            self('after_cancel_train')
        finally:
            self('after_fit')
            # remove callbacks passed with fit,
            # as well as TrainEvalCallback and AvgStatsCallback
            self.remove_callbacks(fit_callbacks)

    def _get_preds(self, valid_preds=True, new_dl=None, record_preds=False):
        """
        The record_preds part still needs work. Ideally, the post-processing of
        the predictions would be done by a callback, and all post-processing
        would be done in the after_batch method of the callback rather than
        hardcoding different cases here (current_Y_hat for regression,
        current_Y_hat for classification etc.).
        """
        # add GetPredictionsCallback
        get_preds_callbacks = [GetPredictionsCallback(listify(self.metrics))]
        self.add_callbacks(get_preds_callbacks)
        # (re)set various Learner attributes, including correct dataloader
        self.loss = tensor(0.)
        self.pred_df = pd.DataFrame(columns=[#'X',
                                             'Y', 'Y_hat'])
        if new_dl == None and valid_preds:
            self.dl = self.data.valid_dl
        elif new_dl == None and not valid_preds:
            self.dl = self.data.train_dl
        elif isinstance(new_dl, DataLoader):
            self.dl = new_dl
        else:
            raise CancelPredsException("No new dataloader: `new_dl` is not the right type.")

        self('begin_get_preds')
        self.iters = len(self.dl)
        with torch.no_grad():
            for i, (xb, yb) in enumerate(self.dl):
                self.iter = i
                self.xb, self.yb = xb, yb;                      self('begin_batch')
                self.pred = self.model(self.xb);                self('after_pred')
                self.loss = self.loss_func(self.pred, self.yb); self('after_loss')
                # record actual predictions if record_preds is True
                if record_preds:
                    current_Y = self.yb.cpu().numpy().squeeze()
                    current_Y_hat = self.pred.cpu().numpy().squeeze()
                    # we need this for classification: Y_hat is a tensor of logits
                    current_Y_hat = current_Y_hat if len(current_Y_hat.shape)<=1 else \
                                    np.argmax(current_Y_hat, axis=1)
                    current_preds_df = pd.DataFrame({#'X':self.xb.cpu().numpy(),
                                                    'Y':current_Y,
                                                    'Y_hat':current_Y_hat},
                                                    index=range(len(current_Y)))
                    self.pred_df = pd.concat([self.pred_df, current_preds_df],
                                            ignore_index=True)
                self('after_batch')
        self('end_get_preds')
        self('after_get_preds')

        self.remove_callbacks(get_preds_callbacks)

    def get_valid_preds(self, record_preds=False):
        return self._get_preds(valid_preds=True, record_preds=record_preds)

    def get_train_preds(self, record_preds=False):
        return self._get_preds(valid_preds=False, record_preds=record_preds)

    def get_new_preds(self, new_dl, record_preds=False):
        return self._get_preds(new_dl=new_dl, record_preds=record_preds)

    def __call__(self, callback_name):
        results = False
        assert callback_name in self.ALL_CALLBACKS
        for callback in sorted(self.callbacks, key=lambda x: x._order):
            results = callback(callback_name) and results
        return results



from torch import nn
# from torch.nn import init
from functools import partial
from .data_block import ListContainer


def children(m): return list(m.children())

def find_modules(m, cond):
    if cond(m): return [m]
    return sum([find_modules(o,cond) for o in m.children()], [])

def is_lin_layer(l):
    lin_layers = (nn.Conv1d, nn.Conv2d, nn.Conv3d, nn.Linear)
    return isinstance(l, lin_layers)

def get_batch(dl, learn):
    learn.xb, learn.yb = next(iter(dl))
    learn.do_begin_fit(0)
    learn('begin_batch')
    learn('after_fit')
    return learn.xb, learn.yb

def model_summary(learn, data, find_all=False, print_mod=True):
    xb, yb = get_batch(data.valid_dl, learn)
    mods = find_modules(learn.model, is_lin_layer) if find_all else learn.model.children()
    f = lambda hook,mod,inp,out: print(f"====\n{mod}\n" if print_mod else "", out.shape)
    with Hooks(mods, f) as hooks:
        learn.model(xb)

def append_stat(hook, mod, inp, outp):
    d = outp.data
    hook.mean,hook.std = d.mean().item(),d.std().item()

def lsuv_module(m, xb):
    """ Usage:
    mods = find_modules(learn.model, is_lin_layer)
    for m in mods: print(lsuv_module(m, xb))"""
    h = Hook(m, append_stat)
    if getattr(m, 'bias', None) is not None:
        while mdl(xb) is not None and abs(h.mean) > 1e-3:
            m.bias.data -= h.mean
    while mdl(xb) is not None and abs(h.std-1) > 1e-3:
        m.weight.data /= h.std
    h.remove()
    return h.mean, h.std


class Hook():
    def __init__(self, m, f): self.hook = m.register_forward_hook(partial(f, self))
    def remove(self): self.hook.remove()
    def __del__(self): self.remove()

def append_stats(hook, mod, inp, outp):
    if not hasattr(hook,'stats'): hook.stats = ([],[])
    means,stds = hook.stats
    if mod.training:
        means.append(outp.data.mean())
        stds .append(outp.data.std())

class Hooks(ListContainer):
    def __init__(self, ms, f):
        super().__init__([Hook(m, f) for m in ms])

    def __enter__(self, *args):
        return self

    def __exit__ (self, *args):
        self.remove()

    def __del__(self):
        self.remove()

    def __delitem__(self, i):
        self[i].remove()
        super().__delitem__(i)

    def remove(self):
        for h in self:
            h.remove()




# mostly copies the code in the datasets module of the fastai v1 library (as of early Oct. 2019)

import os
import requests
import hashlib
import tarfile
from pathlib import Path
from fastprogress.fastprogress import progress_bar

Path.ls = lambda x: list(x.iterdir())

MODEL_URL = 'http://files.fast.ai/models/'
URL = 'http://files.fast.ai/data/examples/'

class URLs():
    "Global constants for dataset and model URLs."
    LOCAL_PATH = Path(os.getcwd())
    S3 = 'https://s3.amazonaws.com/fast-ai-'

    S3_IMAGE    = f'{S3}imageclas/'
    S3_NLP      = f'{S3}nlp/'
    S3_MODEL    = f'{S3}modelzoo/'

    # main datasets
    HUMAN_NUMBERS       = f'{URL}human_numbers'
    IMDB                = f'{S3_NLP}imdb'
    IMDB_SAMPLE         = f'{URL}imdb_sample'
    ML_SAMPLE           = f'{URL}movie_lens_sample'
    MNIST_SAMPLE        = f'{URL}mnist_sample'
    MNIST_TINY          = f'{URL}mnist_tiny'
    MNIST_VAR_SIZE_TINY = f'{S3_IMAGE}mnist_var_size_tiny'

    # image classification datasets
    MNIST        = f'{S3_IMAGE}mnist_png'

    # NLP datasets
    AG_NEWS                 = f'{S3_NLP}ag_news_csv'
    AMAZON_REVIEWS          = f'{S3_NLP}amazon_review_full_csv'
    AMAZON_REVIEWS_POLARITY = f'{S3_NLP}amazon_review_polarity_csv'
    DBPEDIA                 = f'{S3_NLP}dbpedia_csv'
    MT_ENG_FRA              = f'{S3_NLP}giga-fren'
    SOGOU_NEWS              = f'{S3_NLP}sogou_news_csv'
    WIKITEXT                = f'{S3_NLP}wikitext-103'
    WIKITEXT_TINY           = f'{S3_NLP}wikitext-2'
    YAHOO_ANSWERS           = f'{S3_NLP}yahoo_answers_csv'
    YELP_REVIEWS            = f'{S3_NLP}yelp_review_full_csv'
    YELP_REVIEWS_POLARITY   = f'{S3_NLP}yelp_review_polarity_csv'

    #Pretrained models
    OPENAI_TRANSFORMER = f'{S3_MODEL}transformer'
    WT103_FWD          = f'{S3_MODEL}wt103-fwd'
    WT103_BWD          = f'{S3_MODEL}wt103-bwd'

_checks = {
    URLs.AG_NEWS:(11784419, 'b86f328f4dbd072486591cb7a5644dcd'),
    URLs.AMAZON_REVIEWS_POLARITY:(688339454, '676f7e5208ec343c8274b4bb085bc938'),
    URLs.AMAZON_REVIEWS:(643695014, '4a1196cf0adaea22f4bc3f592cddde90'),
    URLs.DBPEDIA:(68341743, '239c7837b9e79db34486f3de6a00e38e'),
    URLs.HUMAN_NUMBERS:(30252, '8a19c3bfa2bcb08cd787e741261f3ea2'),
    URLs.IMDB:(144440600, '90f9b1c4ff43a90d67553c9240dc0249'),
    URLs.IMDB_SAMPLE:(571827, '0842e61a9867caa2e6fbdb14fa703d61'),
    URLs.ML_SAMPLE:(51790, '10961384dfe7c5181460390a460c1f77'),
    URLs.MNIST:(15683414, '03639f83c4e3d19e0a3a53a8a997c487'),
    URLs.MNIST_SAMPLE:(3214948, '2dbc7ec6f9259b583af0072c55816a88'),
    URLs.MNIST_TINY:(342207, '56143e8f24db90d925d82a5a74141875'),
    URLs.MNIST_VAR_SIZE_TINY:(565372, 'b71a930f4eb744a4a143a6c7ff7ed67f'),
    URLs.MT_ENG_FRA:(2598183296, '69573f58e2c850b90f2f954077041d8c'),
    URLs.OPENAI_TRANSFORMER:(432848315, '024b0d2203ebb0cd1fc64b27cf8af18e'),
    URLs.SOGOU_NEWS:(384269937, '950f1366d33be52f5b944f8a8b680902'),
    URLs.WIKITEXT:(190200704, '2dd8cf8693b3d27e9c8f0a7df054b2c7'),
    URLs.WIKITEXT_TINY:(4070055, '2a82d47a7b85c8b6a8e068dc4c1d37e7'),
    URLs.WT103_FWD:(105067061, '7d1114cd9684bf9d1ca3c9f6a54da6f9'),
    URLs.WT103_BWD:(105205312, '20b06f5830fd5a891d21044c28d3097f'),
    URLs.YAHOO_ANSWERS:(319476345, '0632a0d236ef3a529c0fa4429b339f68'),
    URLs.YELP_REVIEWS_POLARITY:(166373201, '48c8451c1ad30472334d856b5d294807'),
    URLs.YELP_REVIEWS:(196146755, '1efd84215ea3e30d90e4c33764b889db'),
}

def _expand_path(fpath): return Path(fpath).expanduser()
def url2name(url): return url.split('/')[-1]

def ifnone(a, b):
    "`a` if `a` is not None, otherwise `b`."
    return b if a is None else a

def download_url(url, dest, overwrite=False, pbar=None,
        show_progress=True, chunk_size=1024*1024, timeout=4, retries=5):
    "Download `url` to `dest` unless it exists and not `overwrite`."
    if os.path.exists(dest) and not overwrite: return

    s = requests.Session()
    s.mount('http://',requests.adapters.HTTPAdapter(max_retries=retries))
    u = s.get(url, stream=True, timeout=timeout)
    try: file_size = int(u.headers["Content-Length"])
    except: show_progress = False

    with open(dest, 'wb') as f:
        nbytes = 0
        if show_progress: pbar = progress_bar(range(file_size), auto_update=False, leave=False, parent=pbar)
        try:
            for chunk in u.iter_content(chunk_size=chunk_size):
                nbytes += len(chunk)
                if show_progress: pbar.update(nbytes)
                f.write(chunk)
        except requests.exceptions.ConnectionError as e:
            fname = url.split('/')[-1]
            from fastai.datasets import Config
            data_dir = Config().data_path()
            timeout_txt =(f'\n Download of {url} has failed after {retries} retries\n'
                          f' Fix the download manually:\n'
                          f'$ mkdir -p {data_dir}\n'
                          f'$ cd {data_dir}\n'
                          f'$ wget -c {url}\n'
                          f'$ tar -zxvf {fname}\n\n'
                          f'And re-run your code once the download is successful\n')
            print(timeout_txt)
            import sys;sys.exit(1)


def url2path(url, data=True, ext='.tgz'):
    "Change `url` to a path."
    name = url2name(url)
    return datapath4file(name, ext=ext, archive=False) if data else modelpath4file(name, ext=ext)

def _url2tgz(url, data=True, ext='.tgz'):
    return datapath4file(f'{url2name(url)}{ext}', ext=ext) if data else modelpath4file(f'{url2name(url)}{ext}', ext=ext)

def modelpath4file(filename, ext='.tgz'):
    "Return model path to `filename`, checking locally first then in the config file."
    local_path = URLs.LOCAL_PATH/'models'/filename
    return local_path

def datapath4file(filename, ext='.tgz', archive=True):
    "Return data path to `filename`, checking locally first then in the config file."
    local_path = URLs.LOCAL_PATH/'data'/filename
    return local_path

def download_data(url, fname=None, data=True, ext='.tgz'):
    "Download `url` to destination `fname`."
    fname = Path(ifnone(fname, _url2tgz(url, data, ext=ext)))
    os.makedirs(fname.parent, exist_ok=True)
    if not fname.exists():
        print(f'Downloading {url}')
        download_url(f'{url}{ext}', fname)
    return fname

def _check_file(fname):
    size = os.path.getsize(fname)
    with open(fname, "rb") as f:
        hash_nb = hashlib.md5(f.read(2**20)).hexdigest()
    return size,hash_nb

def untar_data(url, fname=None, dest=None, data=True, force_download=False):
    "Download `url` to `fname` if `dest` doesn't exist, and un-tgz to folder `dest`."
    dest = url2path(url, data) if dest is None else Path(dest)/url2name(url)
    fname = Path(ifnone(fname, _url2tgz(url, data)))
    if force_download or (fname.exists() and url in _checks and _check_file(fname) != _checks[url]):
        print(f"A new version of the {'dataset' if data else 'model'} is available.")
        if fname.exists(): os.remove(fname)
        if dest.exists(): shutil.rmtree(dest)
    if not dest.exists():
        fname = download_data(url, fname=fname, data=data)
        if url in _checks:
            assert _check_file(fname) == _checks[url], f"Downloaded file {fname} does not match checksum expected! Remove that file from {Config().data_archive_path()} and try your code again."
        tarfile.open(fname, 'r:gz').extractall(dest.parent)
    return dest



import os
import re
import time
import torch
from torch import nn
import matplotlib.pyplot as plt
from functools import partial
from fastprogress import master_bar
from fastprogress import progress_bar
from fastprogress.fastprogress import format_time
from IPython.core.debugger import set_trace
from pathlib import Path
from types import SimpleNamespace
from .datasets import URLs
from .utils import ALL_CALLBACKS, camel2snake, listify, cos_1cycle_anneal, combine_scheds, create_phases
from .utils import CancelTrainException


# this enables us to plot tensors
torch.Tensor.ndim = property(lambda x: len(x.shape))

# to be used with the DebugCallback
callback_types = SimpleNamespace(**{o:o for o in ALL_CALLBACKS})

def print_details(o):
    "print function to be used with the DebugCallback"
    print('parameter groups: ', len(o.opt.param_groups), '\n',
           'hyperparam dicts:', '\n',
           '\n'.join([('dict ' + str(i) + ': ' + str(hyper))
                      for i, hyper in enumerate(o.opt.hypers)]))
    raise CancelTrainException()

class Callback():
    """
    Base Callback class, can be called like a function with attr names as args,
    where attrs are methods (e.g., begin_fit, after_batch etc.) to be called at
    relevant points in the train-eval loop.
    The learn (i.e., learner) argument is the learner, which will incorporate the
    callback as one of its attributes.
    _order keeps track of the priority of the callbacks so that they can be
    called in order of priority.
    __getattr__ percolates attrs from inside the learn object up to the callback.
    The name property converts the callback name from camel to snake; the name
    is then percolated up to the learner and becomes one of its attributes.
    """
    _order = 0
    def set_learner(self, learn):
        self.learn = learn
    def __getattr__(self, k):
        return getattr(self.learn, k)

    @property
    def name(self):
        name = re.sub(r'Callback$', '', self.__class__.__name__)
        return camel2snake(name or 'callback')

    def __call__(self, callback_name):
        f = getattr(self, callback_name, None)
        if f and f():
            return True
        return False

class TrainEvalCallback(Callback):
    """
    Callback to do basic training and evaluation.
    """
    def begin_fit(self):
        self.learn.n_epochs = 0.
        self.learn.n_iter = 0

    def after_batch(self):
        if not self.in_train:
            return
        self.learn.n_epochs += 1./self.iters
        self.learn.n_iter   += 1

    def begin_epoch(self):
        self.learn.n_epochs = self.epoch
        self.model.train()
        self.learn.in_train = True

    def begin_validate(self):
        self.model.eval()
        self.learn.in_train = False

class AvgStats():
    def __init__(self, metrics, in_train):
        self.metrics = listify(metrics)
        self.in_train = in_train
        # dictionary to record metrics for plotting
        self.metrics_dict = {'loss':[]}
        for metric in self.metrics:
            self.metrics_dict[metric.__name__] = []

    def reset(self):
        self.tot_loss = 0.
        self.count = 0
        self.tot_metrics = [0.] * len(self.metrics)

    @property
    def all_stats(self):
        return [self.tot_loss.item()] + self.tot_metrics
    @property
    def avg_stats(self):
        return [statistic/self.count for statistic in self.all_stats]

    def __repr__(self):
        if not self.count: return ""
        return f"{'train' if self.in_train else 'valid'}: {self.avg_stats}"

    def accumulate(self, learn):
        batch_size = learn.xb.shape[0]
        self.tot_loss += learn.loss * batch_size
        self.metrics_dict['loss'].append(learn.loss.detach().cpu().item())
        self.count += batch_size
        for i, metric in enumerate(self.metrics):
            current_statistic = metric(learn.pred, learn.yb)
            self.tot_metrics[i] += current_statistic * batch_size
            self.metrics_dict[metric.__name__].append(current_statistic.detach().cpu().item())

class AvgStatsCallback(Callback):
    # _order = -1
    def __init__(self, metrics):
        self.train_stats = AvgStats(metrics, True)
        self.valid_stats = AvgStats(metrics, False)

    def begin_fit(self):
        met_names = ['loss'] + [m.__name__ for m in self.train_stats.metrics]
        names = ['epoch'] + [f'train_{n}' for n in met_names] + [
            f'valid_{n}' for n in met_names] + ['time']
        self.logger(names)

    def begin_epoch(self):
        self.train_stats.reset()
        self.valid_stats.reset()
        self.start_time = time.time()

    def after_loss(self):
        stats = self.train_stats if self.in_train else self.valid_stats
        with torch.no_grad():
            stats.accumulate(self.learn)

    def after_epoch(self):
        stats = [str(self.epoch)]
        for stat in [self.train_stats, self.valid_stats]:
            stats += [f'{v:.6f}' for v in stat.avg_stats]
        stats += [format_time(time.time() - self.start_time)]
        # We use the logger function of the `Learner` here, which
        # can be customized to write in a file or in a progress bar
        self.logger(stats)

    def after_fit(self):
        """
        Saves the train and valid metric_dicts to the learner for plotting.
        """
        self.learn.train_metrics_dict = self.train_stats.metrics_dict
        self.learn.valid_metrics_dict = self.valid_stats.metrics_dict

class ProgressCallback(Callback):
    _order=-1
    def begin_fit(self):
        # set master bar (for epochs) when we start fitting
        self.master_bar = master_bar(range(self.epochs))
        # initialize master_bar
        self.master_bar.on_iter_begin()
        # replace logger (default: printing) with master_bar
        self.learn.logger = partial(self.master_bar.write, table=True)

    def after_fit(self):
        # tell master_bar we're done
        self.master_bar.on_iter_end()

    def after_batch(self):
        # update progress bar after every batch
        self.progress_bar.update(self.iter)

    def begin_epoch(self):
        # set progress bar at the beginning of every training epoch
        self.set_progress_bar()

    def begin_validate(self):
        # set progress bar at the beginning of validation
        self.set_progress_bar()

    def begin_get_preds(self):
        self.master_bar = master_bar(range(1))
        self.master_bar.on_iter_begin()
        self.learn.logger = partial(self.master_bar.write, table=True)
        self.progress_bar = progress_bar(self.dl, parent=self.master_bar,
                                         auto_update=False)

    def after_get_preds(self):
        self.master_bar.on_iter_end()

    def set_progress_bar(self):
        self.progress_bar = progress_bar(self.dl, parent=self.master_bar,
                                         auto_update=False)
        self.master_bar.update(self.epoch)

class GetPredictionsCallback(Callback):
    """
    Computes the average metrics (including loss) for a DataLoader given the
    current model. The model is in eval mode. Also provides per-batch metrics.
    The actual predictions are recorded directly in Learner._get_preds.
    """
    _order = -1
    def __init__(self, metrics):
        # initialize metrics to be computed; False means we're in validation mode
        self.pred_stats = AvgStats(metrics, False)

    def begin_get_preds(self):
        # set up progress bar related things
        met_names = ['loss'] + [m.__name__ for m in self.pred_stats.metrics] + ['time']
        self.logger(met_names)

        # initialize metrics and time
        self.pred_stats.reset()
        self.start_time = time.time()

        # initialize counters
        self.learn.n_epochs = 0.
        self.learn.n_iter = 0

        # make sure the model is
        self.model.eval()
        self.learn.in_train = False

    def after_loss(self):
        # compute per-batch and current total metrics
        with torch.no_grad():
            self.pred_stats.accumulate(self.learn)

    def after_batch(self):
        # update counters
        self.learn.n_epochs += 1./self.iters
        self.learn.n_iter += 1

    def end_get_preds(self):
        # final update to progress bar
        stats = [f'{v:.6f}' for v in self.pred_stats.avg_stats]
        stats += [format_time(time.time() - self.start_time)]
        self.logger(stats)

        # save the per-batch and total metrics
        self.learn.pred_metrics_dict = self.pred_stats.metrics_dict
        self.learn.pred_avg_stats = {name:(v.cpu().item() if hasattr(v, 'cpu') else v)
                                     for name, v in zip(self.pred_stats.metrics_dict.keys(),
                                                        self.pred_stats.avg_stats)}

class TestCallback(Callback):
    """
    Callback to quickly test the model works. Has to follow TrainEvalCallback.
    """
    _order = 1
    def after_step(self):
        if self.n_iter>=10:
            print("ran {} batches successfully".format(self.n_iter))
            raise CancelTrainException()

class SendToDeviceCallback(Callback):
    def begin_fit(self):
        self.model.to(self.learn.device)

    def begin_batch(self):
        self.learn.xb = self.xb.to(self.learn.device)
        self.learn.yb = self.yb.to(self.learn.device)

class Plotter(Callback):
    """
    Recording functions for everything but the lr are relegated to
    AvgStatsCallback, which in turn leverages the metrics_dict in AvgStats.
    The Plotter callback additional records lr and provides plotting
    functionality.
    """
    _order = 0
    def begin_fit(self):
        self.lrs = []

    def after_batch(self):
        if not self.in_train: return
        # plot the lr for the last layer group
        self.lrs.append(self.opt.hypers[-1]['lr'])

    def plot_lr(self):
        plt.figure(figsize=(6, 4))
        plt.plot(self.lrs)
        plt.xlabel('Iteration / Batch')
        plt.ylabel('Learning Rate')
        plt.tight_layout()

    def plot_lr_finder(self, skip_last=0):
        losses = self.train_metrics_dict['loss']
        n = len(losses)-skip_last
        plt.figure(figsize=(6, 4))
        plt.xscale('log')
        plt.plot(self.lrs[:n], losses[:n])
        plt.xlabel('Log Learning Rate')
        plt.ylabel('Loss')
        plt.tight_layout()

    def _plot_stats(self, stats):
        metrics = stats.keys()
        ncols = 2 if len(metrics) > 1 else 1
        if len(metrics) % ncols == 0:
            nrows = len(metrics) // ncols
        else:
            nrows = len(metrics) // ncols + 1
        fig, axes = plt.subplots(nrows=nrows, ncols=ncols,
                                 figsize=(ncols*6, nrows*4))
        for ax, metric in zip(listify(axes), metrics):
            ax.plot(stats[metric])
            ax.set_xlabel('Iteration / Batch')
            ax.set_ylabel(metric.title())
        plt.tight_layout()

    def plot_train_stats(self):
        self._plot_stats(self.train_metrics_dict)

    def plot_valid_stats(self):
        self._plot_stats(self.valid_metrics_dict)

class LRFinder(Callback):
    _order = 1
    def __init__(self, max_iter=100, min_lr=1e-6, max_lr=10):
        self.max_iter = max_iter
        self.min_lr = min_lr
        self.max_lr = max_lr
        self.best_loss = 1e9

    def begin_fit(self):
        "Saves current weights before we disturb them with lr exploration."
        torch.save(self.model.state_dict(),
                   URLs.LOCAL_PATH/'lr_finder_tmp.pth')

    def begin_batch(self):
        if not self.in_train:
            return
        pos = self.n_iter/self.max_iter
        lr = self.min_lr * (self.max_lr/self.min_lr) ** pos
        for param_group in self.opt.hypers:
            param_group['lr'] = lr

    def after_step(self):
        if (self.n_iter >= self.max_iter) or (self.loss > self.best_loss * 10):
            raise CancelTrainException()
        if self.loss < self.best_loss:
            self.best_loss = self.loss

    def after_fit(self):
        """
        Resets the model weights to what they were before lr exploration.
        Resets optimizer.
        If model or callbacks have a reset method, we call it.
        """
        self.plotter.plot_lr_finder(3)
        self.learn.model.load_state_dict(torch.load(URLs.LOCAL_PATH/'lr_finder_tmp.pth'))
        self.learn.reset_opt = True # reset optimizer after LRFinder
        # future proofing in case we add reset methods to models and callbacks
        if hasattr(self.learn.model, 'reset'):
            self.learn.model.reset()
        for callback in self.callbacks:
            if hasattr(callback, 'reset'): callback.reset()

class GradientClipping(Callback):
    """
    Checks after the backward pass if the norm (sum of squares) of the
    gradients is greater than the number clip; if they are, they get
    divided (scaled down) so that they're smaller than clip.
    """
    def __init__(self, clip=None):
        self.clip = clip

    def after_backward(self):
        if self.clip:
            nn.utils.clip_grad_norm_(self.learn.model.parameters(), self.clip)

class ParamScheduler(Callback):
    """
    Assumes there are as many scheduling functions as there are param groups
    (if only one function is passed, it is cloned for each param group).
    The scheduling functions are then used to set the value of the parameter
    param_name in each param group based on the current position in the training loop.
    """
    _order = 1
    def __init__(self, param_name, sched_funcs):
        self.param_name = param_name
        self.sched_funcs = listify(sched_funcs)

    def begin_batch(self):
        if not self.in_train:
            return
        funcs = self.sched_funcs
        if len(funcs) == 1:
            funcs = funcs * len(self.opt.param_groups)
        pos = self.n_epochs/self.epochs
        for func, hyper in zip(funcs, self.opt.hypers):
            hyper[self.param_name] = func(pos)

def sched_1cycle(lrs, pct_start=0.3, mom_start=0.95, mom_mid=0.85, mom_end=0.95):
    lrs = listify(lrs)
    phases = create_phases(pct_start)
    sched_lr = [combine_scheds(phases, cos_1cycle_anneal(lr*1e-1, lr, lr*1e-5))
                for lr in lrs]
    sched_mom = combine_scheds(phases, cos_1cycle_anneal(mom_start, mom_mid, mom_end))
    return [ParamScheduler('lr', sched_lr), ParamScheduler('mom', sched_mom)]

class RNNTrainer(Callback):
    """
    Adds two L2 penalties on activations (not weights).
    Activation Regularization (AR): ensures activations are not too high.
    Temporal Activation Regularization (TAR): ensures activations don't change
    radically from timestep to timestep.
    """
    def __init__(self, alpha, beta):
        # parameter for Activation Regularization (AR)
        self.alpha = alpha
        # parameter for Temporal Activation Regularization (TAR)
        self.beta = beta

    def after_pred(self):
        # Saves the extra outputs for later and only returns the true output.
        self.raw_out = self.pred[1]
        self.out = self.pred[2]
        self.learn.pred = self.pred[0]

    def after_loss(self):
        # Activation Regularization (AR): we add to the loss an L2 penalty
        # on the last activations of the AWD LSTM (with dropout applied)
        if self.alpha != 0.:
            self.learn.loss += self.alpha * self.out[-1].float().pow(2).mean()
        # Temporal Activation Regularization (TAR): we add to the loss an L2
        # penalty on the difference between two consecutive raw outputs
        # (consecutive in terms of words)
        if self.beta != 0.:
            h = self.raw_out[-1]
            if len(h) > 1:
                self.learn.loss += self.beta * (h[:,1:] - h[:,:-1]).float().pow(2).mean()

    def begin_epoch(self):
        # Shuffle the texts at the beginning of the epoch
        if hasattr(self.dl.dataset, "batchify"):
            self.dl.dataset.batchify()

class DebugCallback(Callback):
    """
    Overrides __call__ itself to ensure we debug callback callback_name.
    When callback_name is called, it call the function func if it's provided,
    otherwise, it calls the debugger (set_trace).
    example usage: callbacks = [DebugCallback(callback_types.after_batch, print_details)]
    """
    _order = 999
    def __init__(self, callback_name, func=None):
        self.callback_name = callback_name
        self.func = func

    def __call__(self, callback_name):
        if callback_name == self.callback_name:
            if self.func:
                self.func(self.learn)
            else:
                set_trace()

class SaveModelCallback(Callback):
    def __init__(self, model_name, path_str=None):
        self.model_name = model_name
        if path_str and isinstance(path_str, str):
            self.abs_path = URLs.LOCAL_PATH/path_str
        else:
            self.abs_path = URLs.LOCAL_PATH

    def after_epoch(self):
        torch.save(self.model.state_dict(),
                   self.abs_path/(self.model_name + '_' + str(int(self.n_epochs-1)) + '.pth'))

    def after_fit(self):
        torch.save(self.model.state_dict(),
                   self.abs_path/(self.model_name + '_final.pth'))



import html
import os
import numpy as np
import re
import spacy
import torch
from collections import defaultdict, Counter
from concurrent.futures import ProcessPoolExecutor
from fastprogress import progress_bar
# from functools import partial
from pathlib import Path
from spacy.symbols import ORTH
from torch import tensor, LongTensor
from torch.utils.data import Sampler, Dataset, DataLoader
from .utils import compose, listify, uniqueify, read_file, sched_exp, get_files, df_names_to_idx


# general data block utility functions

def parent_labeler(filename):
    return filename.parent.name

def get_dls(train_ds, valid_ds, batch_size, **kwargs):
    return (DataLoader(train_ds, batch_size=batch_size, shuffle=True, **kwargs),
            DataLoader(valid_ds, batch_size=batch_size*2, **kwargs))

def grandparent_splitter(filename, valid_name='valid', train_name='train'):
    grandparent = filename.parent.parent.name
    return True if grandparent==valid_name else \
           False if grandparent==train_name else \
           None

def split_by_func(items, f):
    mask = [f(item) for item in items]
    # `None` values will be filtered out
    false_items = [item for item, m in zip(items, mask) if m==False]
    true_items = [item for item, m in zip(items, mask) if m==True]
    return false_items, true_items

def label_by_func(splitdata, func, processor_x=None, processor_y=None):
    train = LabeledData.label_by_func(splitdata.train, func,
                                      processor_x=processor_x, processor_y=processor_y)
    valid = LabeledData.label_by_func(splitdata.valid, func,
                                      processor_x=processor_x, processor_y=processor_y)
    return SplitData(train, valid)

def parallel(func, arr, max_workers=16):
    if max_workers < 2:
        results = list(progress_bar(map(func, enumerate(arr)), total=len(arr)))
    else:
        with ProcessPoolExecutor(max_workers=max_workers) as ex:
            return list(progress_bar(ex.map(func, enumerate(arr)), total=len(arr)))
    if any([result is not None for result in results]):
        return results

# utility functions for text

#special tokens
UNK, PAD, BOS, EOS, TK_REP, TK_WREP, TK_ALLCAPS, TK_CAP = "_UNK_ _PAD_ _BOS_ _EOS_ _REP_ _WREP_ _ALLCAPS_ _CAP_".split()
default_spec_tok = [UNK, PAD, BOS, EOS, TK_REP, TK_WREP, TK_ALLCAPS, TK_CAP]

def sub_br(text):
    "Replaces the <br /> by \n"
    re_br = re.compile(r'<\s*br\s*/?>', re.IGNORECASE)
    return re_br.sub("\n", text)

def spec_add_spaces(text):
    "Add spaces around / and #"
    return re.sub(r'([/#])', r' \1 ', text)

def rm_useless_spaces(text):
    "Remove multiple spaces"
    return re.sub(' {2,}', ' ', text)

def replace_rep(text):
    "Replace repetitions at the character level: cccc -> TK_REP 4 c"
    def _replace_rep(match):
        character, characters = match.groups()
        return f' {TK_REP} {len(characters)+1} {character} '
    re_rep = re.compile(r'(\S)(\1{3,})')
    return re_rep.sub(_replace_rep, text)

def replace_wrep(text):
    "Replace word repetitions: word word word -> TK_WREP 3 word"
    def _replace_wrep(match):
        word, words = match.groups()
        return f' {TK_WREP} {len(words.split())+1} {word} '
    re_wrep = re.compile(r'(\b\w+\W+)(\1{3,})')
    return re_wrep.sub(_replace_wrep, text)

def fixup_text(text):
    "Various messy things we've seen in documents"
    re1 = re.compile(r'  +')
    text = text.replace('#39;', "'").replace('amp;', '&').replace('#146;', "'").replace(
        'nbsp;', ' ').replace('#36;', '$').replace('\\n', "\n").replace('quot;', "'").replace(
        '<br />', "\n").replace('\\"', '"').replace('<unk>',UNK).replace(' @.@ ','.').replace(
        ' @-@ ','-').replace('\\', ' \\ ')
    return re1.sub(' ', html.unescape(text))

# rules to apply before tokenization
default_pre_rules = [fixup_text, replace_rep, replace_wrep, spec_add_spaces, rm_useless_spaces, sub_br]

def replace_all_caps(text):
    "Replace tokens in ALL CAPS by their lower version and add `TK_ALLCAPS` before."
    results = []
    for token in text:
        if token not in default_spec_tok and token.isupper() and len(token) > 1:
            results.append(TK_ALLCAPS); results.append(token.lower())
        else: results.append(token)
    return results

def replace_caps(text):
    "Replace all Capitalized tokens in by their lower version and add `TK_CAP` before."
    results = []
    for token in text:
        if token == '':
            continue
        if token not in default_spec_tok and token[0].isupper() and len(token) > 1 and token[1:].islower():
            results.append(TK_CAP)
        results.append(token.lower())
    return results

def add_eos_bos(text):
    "adds 'beginning of string' and 'end of string' tokens"
    # return [BOS] + text + [EOS]
    return [BOS] + text # adding only BOS; EOS seems redundant

# rules to apply after tokenization
default_post_rules = [replace_caps, replace_all_caps, add_eos_bos]

def get_lm_dls(train_ds, valid_ds, batch_size, bptt, **kwargs):
    return (DataLoader(LM_PreLoader(train_ds, batch_size, bptt, shuffle=True),
                       batch_size=batch_size, **kwargs),
            DataLoader(LM_PreLoader(valid_ds, batch_size, bptt, shuffle=False),
                       batch_size=2*batch_size, **kwargs))

def lm_databunchify(splitdata, batch_size, bptt, **kwargs):
    return DataBunch(*get_lm_dls(splitdata.train, splitdata.valid, batch_size, bptt, **kwargs))

def pad_collate(samples, pad_idx=1, pad_first=False):
    # identify the longest document in the minibatch
    max_len = max([len(sample[0]) for sample in samples])
    # create rectangular tensor that can accommodate all documents
    # in the batch up to that max_len, and fill it with padding.
    results = torch.zeros(len(samples), max_len).long() + pad_idx
    # take documents in the minibatch and put them in the tensor
    # keeping padding either at the beginning or at the end
    for i, sample in enumerate(samples):
        if pad_first:
            results[i, -len(sample[0]):] = LongTensor(sample[0])
        else:
            results[i, :len(sample[0]) ] = LongTensor(sample[0])
    return results, tensor([sample[1] for sample in samples])

def get_clas_dls(train_ds, valid_ds, batch_size, **kwargs):
    train_sampler = SortishSampler(train_ds.x,
                                   key=lambda t: len(train_ds.x[t]),
                                   batch_size=batch_size)
    valid_sampler = SortSampler(valid_ds.x,
                                key=lambda t: len(valid_ds.x[t]))
    return (DataLoader(train_ds, batch_size=batch_size, sampler=train_sampler,
                       collate_fn=pad_collate, **kwargs),
            DataLoader(valid_ds, batch_size=batch_size*2, sampler=valid_sampler,
                       collate_fn=pad_collate, **kwargs))

def clas_databunchify(splitdata, batch_size, **kwargs):
    return DataBunch(*get_clas_dls(splitdata.train, splitdata.valid, batch_size, **kwargs))


# classes

class DataBunch():
    """
    Bunches together a train dataloader and a validation dataloader.
    """
    def __init__(self, train_dl, valid_dl, c_in=None, c_out=None):
        self.train_dl = train_dl
        self.valid_dl = valid_dl
        self.c_in = c_in
        self.c_out = c_out

    @property
    def train_ds(self): return self.train_dl.dataset

    @property
    def valid_ds(self): return self.valid_dl.dataset

class ListContainer():
    """
    Base class for indepedents and dependent variables (vectors of observations).
    """
    def __init__(self, items):
        self.items = listify(items)

    def __getitem__(self, idx):
        try:
            return self.items[idx]
        except TypeError:
            if isinstance(idx[0], bool):
                # check that the idx are a boolean mask
                assert len(idx) == len(self)
                return [item for mask, item in zip(idx, self.items) if mask]
            return [self.items[i] for i in idx]

    def __len__(self):
        return len(self.items)

    def __iter__(self):
        return iter(self.items)

    def __setitem__(self, i, o):
        self.items[i] = o

    def __delitem__(self, i):
        del(self.items[i])

    def __repr__(self, n=3):
        results = f'{self.__class__.__name__} ({len(self)} items)\n[{", ".join(str(self.items[i]) for i in range(n))}]'
        if len(self) > n:
            results = results[:-1] + ', ...]'
        return results

class ItemList(ListContainer):
    def __init__(self, items, path='.', transforms=None):
        super().__init__(items)
        self.path = Path(path)
        self.transforms = transforms

    def __repr__(self):
        return f'{super().__repr__()}\nPath: {self.path}'

    def new(self, items, cls=None):
        if cls is None:
            cls=self.__class__
        return cls(items, self.path, transforms=self.transforms)

    def  get(self, i):
        return i

    def _get(self, i):
        return compose(self.get(i), self.transforms)

    def __getitem__(self, idx):
        results = super().__getitem__(idx)
        if isinstance(results, list):
            return [self._get(result) for result in results]
        return self._get(results)

class TextList(ItemList):
    @classmethod
    def from_files(cls, path, extensions='.txt',
                   recurse=True, include=None, **kwargs):
        return cls(get_files(path, extensions, recurse=recurse, include=include),
                   path, **kwargs)

    @classmethod
    def from_df(cls, path, df, col_names, **kwargs):
        "Create an ItemList in path from the inputs in the col_names of df."
        cols = df.iloc[:, df_names_to_idx(col_names, df)]
        assert not cols.isna().any().any(), f"You have NaN values in column(s) {col_names} of your dataframe, please fix it."
        return cls(_maybe_squeeze(cols.values), path, **kwargs)

    def get(self, i):
        if isinstance(i, Path):
            return read_file(i)
        return i

class SplitData():
    def __init__(self, train, valid):
        self.train = train
        self.valid = valid

    def __getattr__(self, k):
        return getattr(self.train, k)

    # this is needed if we want to pickle SplitData and
    # load it back without recursion errors
    def __setstate__(self, data):
        self.__dict__.update(data)

    @classmethod
    def split_by_func(cls, itemlist, func):
        """
        creates a SplitData object by splitting an itemlist intro a train and
        validation set using func, then feeding the 2 sets into SplitData
        """
        lists = map(itemlist.new, split_by_func(itemlist.items, func))
        return cls(*lists)

    def __repr__(self):
        return f'{self.__class__.__name__}\nTrain: {self.train}\nValid: {self.valid}\n'

def _label_by_func(dataset, func, cls=ItemList):
    return cls([func(item) for item in dataset.items], path=dataset.path)

def _label_from_list(dataset, labels, cls=ItemList):
    labels = listify(labels)
    return cls(labels, path=dataset.path)

class LabeledData():
    """
    Associates an itemlist x (the independent variable)
    with an itemlist y of labels (the dependent variable).
    Assumes x and y are itemlists.
    """
    def __init__(self, x, y, processor_x=None, processor_y=None):
        self.x = self.process(x, processor_x)
        self.y = self.process(y, processor_y)
        self.processor_x = processor_x
        self.processor_y = processor_y

    def __repr__(self):
        return f'{self.__class__.__name__}\nx: {self.x}\ny: {self.y}\n'

    def __getitem__(self,idx):
        return self.x[idx], self.y[idx]

    def __len__(self):
        return len(self.x)

    def process(self, itemlist, processor):
        return itemlist.new(compose(itemlist.items, processor))

    def obj(self, items, idx, processors):
        isint = isinstance(idx, int) or \
                (isinstance(idx, torch.LongTensor) and not idx.ndim)
        item = items[idx]
        for processor in reversed(listify(processors)):
            item = processor.deprocess_one(item) if isint else processor.deprocess(item)
        return item

    def x_obj(self, idx):
        return self.obj(self.x, idx, self.processor_x)

    def y_obj(self, idx):
        return self.obj(self.y, idx, self.processor_y)

    @classmethod
    def label_by_func(cls, itemlist, func, processor_x=None, processor_y=None):
        return cls(itemlist, _label_by_func(itemlist, func),
                   processor_x=processor_x, processor_y=processor_y)

    @classmethod
    def label_from_df(cls, itemlist, df, col_names, processor_x=None, processor_y=None):
        labels = df.iloc[:, df_names_to_idx(col_names, df)]
        assert labels.isna().sum().sum() == 0, f"You have NaN values in column(s) {cols} of your dataframe, please fix it."
        return cls(itemlist, _label_from_list(itemlist, _maybe_squeeze(labels)),
                   processor_x=processor_x, processor_y=processor_y)

class Processor():
    def process(self, items):
        return items

class CategoryProcessor(Processor):
    """
    Converts a factor (vector of categorical levels) to an int-based vector.
    """
    def __init__(self):
        self.level_dict = None

    def __call__(self, items):
        # The level_dict which contains all the levels of the factor
        # is defined on the first use.
        if self.level_dict is None:
            self.level_dict = uniqueify(items)
            self.otoi = {level:idx for idx,level in enumerate(self.level_dict)}
        return [self.process_one(item) for item in items]

    def process_one(self, item):
        "process one item"
        return self.otoi[item]

    def deprocess(self, idxs):
        assert self.level_dict is not None
        return [self.deprocess_one(idx) for idx in idxs]

    def deprocess_one(self, idx):
        "deprocess one index"
        return self.level_dict[idx]

class TokenizeProcessor(Processor):
    def __init__(self, lang="en", chunksize=2000, pre_rules=None,
                 post_rules=None, max_workers=8):
        self.chunksize = chunksize
        self.max_workers = max_workers
        self.tokenizer = spacy.blank(lang).tokenizer
        for word in default_spec_tok:
            self.tokenizer.add_special_case(word, [{ORTH: word}])
        self.pre_rules = default_pre_rules if pre_rules is None else pre_rules
        self.post_rules = default_post_rules if post_rules is None else post_rules

    def process(self, chunk):
        i, chunk = chunk # needed because parallel has an enumerate inside
        chunk = [compose(text, self.pre_rules) for text in chunk]
        texts = [[d.text for d in doc] for doc in self.tokenizer.pipe(chunk)]
        texts = [compose(text, self.post_rules) for text in texts]
        return texts

    def __call__(self, items):
        toks = []
        if isinstance(items[0], Path):
            items = [read_file(i) for i in items]
        chunks = [items[i: i+self.chunksize]
                  for i in (range(0, len(items), self.chunksize))]
        toks = parallel(self.process, chunks, max_workers=self.max_workers)
        return sum(toks, [])

    def deprocess(self, toks):
        return [self.deprocess_one(tok) for tok in toks]

    def deprocess_one(self, tok):
        "deprocess one index"
        return " ".join(tok)

class NumericalizeProcessor(Processor):
    def __init__(self, vocab=None, max_vocab=60000, min_freq=2):
        self.vocab = vocab
        self.max_vocab = max_vocab
        self.min_freq = min_freq

    def __call__(self, texts):
        #The vocab is defined on the first use.
        if self.vocab is None:
            freq = Counter(word for text in texts for word in text)
            self.vocab = [word for word,frequency in freq.most_common(self.max_vocab) \
                          if frequency >= self.min_freq]
            for word in reversed(default_spec_tok):
                if word in self.vocab:
                    self.vocab.remove(word)
                self.vocab.insert(0, word)
        if getattr(self, 'otoi', None) is None:
            self.otoi = defaultdict(int, {v:k for k,v in enumerate(self.vocab)})
        return [self.process_one(word) for word in texts]

    def process_one(self, text):
        "process one text"
        return [self.otoi[word] for word in text]

    def deprocess(self, idxs):
        assert self.vocab is not None
        return [self.deprocess_one(idx) for idx in idxs]

    def deprocess_one(self, idx):
        "deprocess one index"
        if isinstance(idx, list):
            return [self.vocab[i] for i in idx]
        return self.vocab[idx]

class LM_PreLoader():
    """
    Creates independent and dependent variables for language models.
    Assumes the data coming in is LabeledData with the texts stored
    in the independent variable slot x.
    We concatenate all texts in a stream, then break it into batch_size
    many texts, which are contiguous pieces of text of length text_len.
    We generate x and y labels (text to be ingested, and to be predicted)
    by tiling every text with adjacent, non-overlapping windows that are
    bptt wide, i.e., have bptt many words in them.
    The sequence of words in each bptt-wide window is the independent variable,
    and the dependent variable is the indep. var. sequence shifted by 1 word.
    See __getitem__ below, where:
    -- indep. var. (text to ingested): text[seq_idx:seq_idx+self.bptt]
    -- dep. var. (text to be predicted): text[seq_idx+1:seq_idx+self.bptt+1]
    """
    def __init__(self, data, batch_size=64, bptt=70, shuffle=False):
        self.data = data
        # batch_size: number of data points (bptt-wide windows of words) per batch
        self.batch_size = batch_size
        # back-propagation through time (bptt), that is, the length of
        # the ingested sequence of words before we predict next word
        self.bptt = bptt
        self.shuffle = shuffle
        # total length of concatenated text in words
        total_len = sum([len(text) for text in data.x])
        # we will break the concatenated text into batch_size
        # contiguous pieces of text, each of which will have length text_len
        self.text_len = total_len // batch_size
        self.batchify()

    def __len__(self):
        """
        We obtain total length by calculating length per text first.
        Each text has text_len words, and we need to count how many adjacent,
        non-overlapping bptt-wide windows we can tile the text with.
        Specifically, we need to tile text_len - 1 words only, because that
        final word can only be part of the dependent variable (it can only be
        predicted, not ingested).
        Once we have the length of a text, which is (self.text_len-1) // self.bptt,
        that is, the number of bptt-wide tiles, we multiply by batch_size, which
        is the number of indepedent bptt windows we'll have in a batch.
        """
        return ((self.text_len-1) // self.bptt) * self.batch_size

    def __getitem__(self, idx):
        """
        idx is an index limited to the total length of the dataset, given by
        __len__ above. Recall that each data point in the dataset consists of
        a bptt-wide window of words that we tiled the text with.
        When we index into the dataset with the index idx, we need to retrieve
        a bptt window of words (text to be ingested), paired with that same
        window shifted by 1 word to the right.


        To see what how the indexation works with a simple example, create a dataset
        of total length 120, set batch_size to 10 and break it into batch_size contiguous
        sequences:

        x = list(range(1, 121))
        batch_size = 10
        x = torch.tensor(x).view(batch_size, -1)

        The dataset x is now:

        tensor([[  1,   2,   3,   4,   5,   6,   7,   8,   9,  10,  11,  12],
                [ 13,  14,  15,  16,  17,  18,  19,  20,  21,  22,  23,  24],
                [ 25,  26,  27,  28,  29,  30,  31,  32,  33,  34,  35,  36],
                [ 37,  38,  39,  40,  41,  42,  43,  44,  45,  46,  47,  48],
                [ 49,  50,  51,  52,  53,  54,  55,  56,  57,  58,  59,  60],
                [ 61,  62,  63,  64,  65,  66,  67,  68,  69,  70,  71,  72],
                [ 73,  74,  75,  76,  77,  78,  79,  80,  81,  82,  83,  84],
                [ 85,  86,  87,  88,  89,  90,  91,  92,  93,  94,  95,  96],
                [ 97,  98,  99, 100, 101, 102, 103, 104, 105, 106, 107, 108],
                [109, 110, 111, 112, 113, 114, 115, 116, 117, 118, 119, 120]])

        Now assume a bptt window of 3. Suppose we want to find the datapoint,
        that is, the 3-word-wide window, at index 23:

        bptt = 3
        idx = 23

        To get the text in that window, we first find the continuous piece of text:

        text = x[idx % batch_size]

        The text is:

        tensor([37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48])

        Then, we find the first word in that text that has not been consumed yet:

        seq_idx = (idx // batch_size) * bptt

        This is word number 6. So the two tensors we return are:

        text[seq_idx:seq_idx+bptt]
        text[seq_idx+1:seq_idx+bptt+1]

        Namely:
         -- tensor([43, 44, 45]) -- window to be ingested
         -- tensor([44, 45, 46]) -- window containing the prediction (the last word)
        """
        # We consume bptt windows from the dataset in batches of batch_size. So,
        # to identify the piece of contiguous text associated with the bptt window
        # at the index idx, we take the remainder idx % self.batch_size, then we
        # get the text that the current bptt window comes from
        text = self.batched_data[idx % self.batch_size]

        # Now we need to find out how much of this contiguous text
        # we already consumed by the time we got to the bptt window at index idx.
        # To find this, we need to compute the number of complete batches we went
        # through up to this point, which is idx // self.batch_size.
        # In each of those batches, we consumed one bptt-wide window of words from
        # the contiguous text, so the total number of consumed words is
        # number of complete batches (idx // self.batch_size) multiplied by bptt.
        seq_idx = (idx // self.batch_size) * self.bptt

        # we now have the bptt window to be ingested, and the window to be predicted
        return text[seq_idx:seq_idx+self.bptt], text[seq_idx+1:seq_idx+self.bptt+1]

    def batchify(self):
        texts = self.data.x
        if self.shuffle:
            texts = texts[torch.randperm(len(texts))]
        stream = torch.cat([tensor(text) for text in texts])
        # each column is a batch that has batch_size words
        # every row is a contiguous piece of text of length text_len
        self.batched_data = stream[:self.text_len * self.batch_size].view(self.batch_size, self.text_len)

class SortSampler(Sampler):
    """
    This looks at all the documents in data_source and sorts them in reverse
    order according to the key that's passed in.
    Example usage:
    See SortishSampler below.
    SortSampler is used for validation, ordering documents deterministically in
    reverse order of length. SortishSampler should be used for training.
    """
    def __init__(self, data_source, key):
        self.data_source = data_source
        self.key = key

    def __len__(self):
        return len(self.data_source)

    def __iter__(self):
        return iter(sorted(list(range(len(self.data_source))),
                           key=self.key, reverse=True))

class SortishSampler(Sampler):
    """
    Sampler that sorts the documents in reverse order with some random
    shuffling. Every batch has documents of a similar length, but with some
    randomness.
    We create random megabatches (50 times bigger than batch_size) and
    sort each of them in reverse order of length.
    Example usage, sorting on the length of the documents:
    SortishSampler(ll.train.x,
                   key=lambda t: len(ll.train[int(t)][0]),
                   batch_size=batch_size)
    """
    def __init__(self, data_source, key, batch_size):
        self.data_source = data_source
        self.key = key
        self.batch_size = batch_size

    def __len__(self):
        return len(self.data_source)

    def __iter__(self):
        idxs = torch.randperm(len(self.data_source))
        megabatches = [idxs[i:i+self.batch_size*50] for i in range(0, len(idxs), self.batch_size*50)]
        sorted_idx = torch.cat([tensor(sorted(s, key=self.key, reverse=True)) for s in megabatches])
        batches = [sorted_idx[i:i+self.batch_size] for i in range(0, len(sorted_idx), self.batch_size)]
        # find the chunk with the largest key
        max_idx = torch.argmax(tensor([self.key(ck[0]) for ck in batches]))
        # then make sure it goes first
        batches[0], batches[max_idx] = batches[max_idx], batches[0]
        batch_idxs = torch.randperm(len(batches)-2)
        sorted_idx = torch.cat([batches[i+1] for i in batch_idxs]) if len(batches) > 1 else LongTensor([])
        sorted_idx = torch.cat([batches[0], sorted_idx, batches[-1]])
        return iter(sorted_idx)




import operator
import os
import re
import math
import numpy as np
import random
import torch
from collections import OrderedDict
from collections.abc import Iterable
from functools import partial
from torch import nn
from torch import tensor
from pathlib import Path


# The inventory of all possible callback functions
ALL_CALLBACKS = {'begin_batch', 'after_pred', 'after_loss', 'after_backward',
                 'after_step', 'after_cancel_batch', 'after_batch',
                 'after_cancel_epoch', 'begin_fit', 'begin_epoch', 'begin_validate',
                 'after_epoch', 'after_cancel_train', 'after_fit', 'begin_get_preds',
                 'end_get_preds', 'after_get_preds'}

# learner exceptions

class CancelTrainException(Exception): pass
class CancelEpochException(Exception): pass
class CancelBatchException(Exception): pass
class CancelPredsException(Exception): pass

# tests

def test(a, b, cmp, cname=None):
    if cname is None: cname = cmp.__name__
    assert cmp(a, b), f"{cname}:\n{a}\n{b}"

def test_eq(a, b): test(a, b, operator.eq, '==')

def near(a, b): return torch.allclose(a, b, rtol=1e-3, atol=1e-5)
def test_near(a, b): test(a, b, near)


# renaming attrs

_camel_re1 = re.compile('(.)([A-Z][a-z]+)')
_camel_re2 = re.compile('([a-z0-9])([A-Z])')

def camel2snake(name):
    s1 = re.sub(_camel_re1, r'\1_\2', name)
    return re.sub(_camel_re2, r'\1_\2', s1).lower()


# lr schedulers

def annealer(f):
    def _inner(start, end): return partial(f, start, end)
    return _inner

@annealer
def sched_cos(start, end, pos):
    return start + (1 + math.cos(math.pi*(1-pos))) * (end-start) / 2

@annealer
def sched_lin(start, end, pos): return start + pos*(end-start)
@annealer
def sched_no(start, end, pos):  return start
@annealer
def sched_exp(start, end, pos): return start * (end/start) ** pos

def cos_1cycle_anneal(start, high, end):
    return [sched_cos(start, high), sched_cos(high, end)]

def combine_scheds(pcts, scheds):
    """
    pcts are the percentages of time the schedules in sched should be
    active.
    """
    assert len(pcts) == len(scheds)
    assert sum(pcts) == 1.
    pcts = tensor([0] + listify(pcts))
    assert torch.all(pcts >= 0)
    pcts = torch.cumsum(pcts, 0)
    def _inner(pos):
        idx = (pos >= pcts).nonzero().max()
        actual_pos = (pos-pcts[idx]) / (pcts[idx+1]-pcts[idx])
        return scheds[idx](actual_pos)
    return _inner

def create_phases(phases):
    phases = listify(phases)
    return phases + [1-sum(phases)]


# data block / dataset related functions

def read_file(filename):
    with open(filename, 'r', encoding = 'utf8') as f: return f.read()

def _get_files(path, filenames, extensions=None):
    path = Path(path)
    results = [path/filename for filename in filenames \
               if not filename.startswith('.') and \
               ((not extensions) or \
                f'.{filename.split(".")[-1].lower()}' in extensions)]
    return results

def get_files(path, extensions=None, recurse=False, include=None):
    path = Path(path)
    extensions = setify(extensions)
    extensions = {extension.lower() for extension in extensions}
    if recurse:
        results = []
        for i, (root, dirs, files) in enumerate(os.walk(path)):
            if include is not None and i == 0:
                dirs[:] = [dir for dir in dirs if dir in include]
            else:
                dirs[:] = [dir for dir in dirs if not dir.startswith('.')]
            results += _get_files(root, files, extensions)
        return results
    else:
        files = [item.name for item in os.scandir(path) if item.is_file()]
        return _get_files(path, files, extensions)

def df_names_to_idx(col_names, df):
    "Return the column indexes of col_names in df."
    if isinstance(col_names[0], int):
        return col_names
    return [df.columns.get_loc(col_name) for col_name in col_names]

def is1d(arraylike_object):
    "Return `True` if `arraylike_object` is one-dimensional"
    return len(arraylike_object.shape) == 1 if hasattr(arraylike_object, 'shape') else len(np.array(arraylike_object).shape) == 1

def _maybe_squeeze(array):
    return (array if is1d(array) else np.squeeze(array))

def match_embeds(old_wgts, old_vocab, new_vocab):
    """
    Matches embeddings from an old_vocab to a new_vocab when transfer learning:
    -- old_vocab is the vocab associated with the pretrained model
    -- new_vocab is the vocab associated with the new corpus
    -- old_wgts are the weights from the old pretrained model (a state dict)
    We end up with embeddings for the new_vocab that are the same as the old
    ones whenever an item is both in the new_vocab and in the old_vocab. When an
    item in the new_vocab is missing from the old_vocab, it is assigned an
    average embedding.
    The old_wgts are updated with respect to the relevant layers. The parameters
    of the other layers are kept the same. The updated old_wgts are returned in
    full so that they can be loaded into the new model.
    """
    wgts = old_wgts['0.emb.weight']
    bias = old_wgts['1.decoder.bias']
    # compute mean weights; we'll assign them to new vocab items
    wgts_m, bias_m = wgts.mean(dim=0), bias.mean()
    # initialize new weights
    new_wgts = wgts.new_zeros(len(new_vocab), wgts.size(1))
    new_bias = bias.new_zeros(len(new_vocab))
    # reverse old vocab so that we can index into the old weights
    otoi = {v:k for k,v in enumerate(old_vocab)}
    # we check every item in the new vocab
    for i,w in enumerate(new_vocab):
        # if the item is in the old_vocab, we transfer the old weights
        if w in otoi:
            idx = otoi[w]
            new_wgts[i], new_bias[i] = wgts[idx], bias[idx]
        # if the item is not in the old_vocab, we give it average weights
        else: new_wgts[i], new_bias[i] = wgts_m, bias_m
    old_wgts['0.emb.weight']        = new_wgts
    old_wgts['0.emb_dropout.emb.weight'] = new_wgts
    old_wgts['1.decoder.weight']    = new_wgts
    old_wgts['1.decoder.bias']      = new_bias
    return old_wgts

# misc

def listify(collection):
    if collection is None:
        return []
    if isinstance(collection, list):
        return collection
    if isinstance(collection, str):
        return [collection]
    if isinstance(collection, Iterable):
        return list(collection)
    return [collection]

def setify(collection):
    return collection if isinstance(collection,set) else \
           set(listify(collection))

def uniqueify(x, sort=False):
    results = list(OrderedDict.fromkeys(x).keys())
    if sort:
        results.sort()
    return results

def compose(x, funcs, *args, order_key='_order', **kwargs):
    """
    Composes the functions in funcs and applies them to x.
    More expplicitly:
    -- we sort the functions in funcs according to order_key
    -- we apply them successively to the input x
    """
    key = lambda o: getattr(o, order_key, 0)
    for f in sorted(listify(funcs), key=key):
        x = f(x, **kwargs)
    return x

def param_getter(m):
    return m.parameters()

def random_splitter(filename, proportion_valid):
    return random.random() < proportion_valid

# def normalize(x, m, s):
    # return (x-m)/s

# def normalize_to(train, valid):
    # m,s = train.mean(),train.std()
    # return normalize(train, m, s), normalize(valid, m, s)

# class Lambda(nn.Module):
    # def __init__(self, func):
        # super().__init__()
        # self.func = func

    # def forward(self, x):
        # return self.func(x)

# def flatten(x):
    # return x.view(x.shape[0], -1)

# class Flatten(nn.Module):
    # def forward(self, x):
        # return x.view(x.size(0), -1)

# def lin_comb(v1, v2, beta): return beta*v1 + (1-beta)*v2

def unsqueeze(input, dims):
    for dim in listify(dims): input = torch.unsqueeze(input, dim)
    return input

def reduce_loss(loss, reduction='mean'):
    return loss.mean() if reduction == 'mean' else \
           loss.sum() if reduction == 'sum' else \
           loss

def set_grad(model, boolean_value):
    "(de)activates gradient for layers in model if not linear or batchnorm"
    if isinstance(model, (nn.Linear, nn.BatchNorm1d)):
        return
    if hasattr(model, 'weight'):
        for param in model.parameters():
            param.requires_grad_(boolean_value)

# def apply_mod(model, func):
    # "applies func to layers in model recursively"
    # func(model)
    # for child in model.children():
        # apply_mod(child, func)

# example usage:
# apply_mod(learn.model, partial(set_grad, boolean_value=False))

# but pytorch already has a version of apply_mod:
# learn.model.apply(partial(set_grad, boolean_value=False))



import torch
import torch.nn.functional as F


def mse(output, targ):
    return (output.squeeze(-1) - targ).pow(2).mean()

def accuracy(out, yb):
    return (torch.argmax(out, dim=1)==yb).float().mean()

def accuracy_thresh(out, yb, thresh=0.5, sigmoid=True):
    "Computes accuracy when `out` and `yb` are the same size."
    if sigmoid:
        out = out.sigmoid()
    return ((out>thresh).byte()==yb.byte()).float().mean()

def cross_entropy_flat(input, target):
    "ensures batch and sequence length dimensions are flattened"
    batch_size, seq_len = target.size()
    return F.cross_entropy(input.view(batch_size * seq_len, -1), target.view(batch_size * seq_len))

def accuracy_flat(input, target):
    "ensures batch and sequence length dimensions are flattened"
    batch_size, seq_len = target.size()
    return accuracy(input.view(batch_size * seq_len, -1), target.view(batch_size * seq_len))



import torch
import warnings
import torch.nn.functional as F
from torch import nn
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence

def dropout_mask(x, size, prob):
    """
    We pass size in so that we get broadcasting along the sequence dimension
    in RNNDropout.
    """
    return x.new(*size).bernoulli_(1-prob).div_(1-prob)

class RNNDropout(nn.Module):
    """
    Note the way size is passed in the forward function: we insert a 3rd
    dimension in between the width and height of the minibatch:
    (x.size(0), 1, x.size(2)).
    The middle dimension is the sequence dimension, so the zeroed-out positions
    will stay the same along the sequence, i.e., througout the bptt sequence.
    """
    def __init__(self, prob=0.5):
        super().__init__()
        self.prob = prob

    def forward(self, x):
        if not self.training or self.prob == 0.:
            return x
        mask = dropout_mask(x.data, (x.size(0), 1, x.size(2)), self.prob)
        return x * mask

class WeightDropout(nn.Module):
    """
    Dropout to the weights (not activations!) of the inner LSTM hidden to
    hidden matrix. We want to preserve the CuDNN speed and not reimplement
    the cell from scratch, so in __init__, we add a parameter that will
    contain the raw weights f'{layer}_raw'. We then replace the weight matrix
    in the LSTM in forward when we call self._setweights()
    """
    def __init__(self, inner_module, weight_prob=[0.], layer_names=['weight_hh_l0']):
        super().__init__()
        self.inner_module = inner_module
        self.weight_prob = weight_prob
        self.layer_names = layer_names
        for layer in self.layer_names:
            # we make a copy of the weights of the selected layers
            weights = getattr(self.inner_module, layer)
            self.register_parameter(f'{layer}_raw', nn.Parameter(weights.data))
            # we apply dropout to the actual weights since we are doing dropout
            # after all, but the forward method will use raw_weights
            self.inner_module._parameters[layer] = F.dropout(weights,
                                                       p=self.weight_prob,
                                                       training=False)

    def _setweights(self):
        "Apply dropout to raw_weights and set them as the layer weights."
        for layer in self.layer_names:
            raw_weights = getattr(self, f'{layer}_raw')
            self.inner_module._parameters[layer] = F.dropout(raw_weights,
                                                       p=self.weight_prob,
                                                       training=self.training)

    def forward(self, *args):
        self._setweights()
        with warnings.catch_warnings():
            #To avoid the warning that comes because the weights aren't flattened.
            warnings.simplefilter("ignore")
            return self.inner_module.forward(*args)

class EmbeddingDropout(nn.Module):
    """
    Applies dropout in the embedding layer by zeroing out some elements of
    the embedding vector. Dropout is applied to full rows of the embedding
    matrix: we drop out entire words and not components of a word's dense
    embedding.
    """
    def __init__(self, emb, embed_prob):
        super().__init__()
        self.emb = emb
        self.embed_prob = embed_prob
        self.pad_idx = self.emb.padding_idx
        if self.pad_idx is None:
            self.pad_idx = -1

    def forward(self, words, scale=None):
        if self.training and self.embed_prob != 0:
            size = (self.emb.weight.size(0), 1)
            mask = dropout_mask(self.emb.weight.data, size, self.embed_prob)
            masked_embed = self.emb.weight * mask
        else:
            masked_embed = self.emb.weight
        if scale:
            masked_embed.mul_(scale)
        return F.embedding(words, masked_embed, self.pad_idx, self.emb.max_norm,
                           self.emb.norm_type, self.emb.scale_grad_by_freq,
                           self.emb.sparse)

def to_detach(h):
    "Detaches h from its gradient history."
    return h.detach() if type(h) == torch.Tensor else tuple(to_detach(v) for v in h)

class AWD_LSTM(nn.Module):
    "AWD-LSTM inspired by https://arxiv.org/abs/1708.02182."
    initrange=0.1

    def __init__(self, vocab_size, emb_size, n_hid, n_layers, pad_token,
                 hidden_prob=0.2, input_prob=0.6, embed_prob=0.1, weight_prob=0.5):
        super().__init__()
        self.batch_size = 1
        self.emb_size = emb_size
        self.n_hid = n_hid
        self.n_layers = n_layers
        self.emb = nn.Embedding(vocab_size, emb_size, padding_idx=pad_token)
        self.emb_dropout = EmbeddingDropout(self.emb, embed_prob)
        # we create n_layers of LSTMs
        self.rnns = [nn.LSTM(emb_size if l == 0 else n_hid,
                             (n_hid if l != n_layers-1 else emb_size),
                             1, batch_first=True)
                     for l in range(n_layers)]
        # we add dropout to the LSTM layers
        self.rnns = nn.ModuleList([WeightDropout(rnn, weight_prob)
                                   for rnn in self.rnns])
        self.emb.weight.data.uniform_(-self.initrange, self.initrange)
        self.input_dropout = RNNDropout(input_prob)
        self.hidden_dropouts = nn.ModuleList([RNNDropout(hidden_prob)
                                              for l in range(n_layers)])

    def forward(self, input):
        batch_size, seq_len = input.size()
        if batch_size != self.batch_size:
            self.batch_size = batch_size
            self.reset()
        raw_output = self.input_dropout(self.emb_dropout(input))
        new_hidden, raw_outputs, outputs = [], [], []
        # we loop through the LSTM layers (plus the hidden dropout layers)
        for l, (rnn, hid_dropout) in enumerate(zip(self.rnns, self.hidden_dropouts)):
            raw_output, new_h = rnn(raw_output, self.hidden[l])
            new_hidden.append(new_h)
            raw_outputs.append(raw_output)
            # we do hidden dropout for all layers but the last one
            if l != self.n_layers - 1: raw_output = hid_dropout(raw_output)
            outputs.append(raw_output)
        self.hidden = to_detach(new_hidden)
        return raw_outputs, outputs

    def _one_hidden(self, l):
        "Return one hidden state."
        nh = self.n_hid if l != self.n_layers - 1 else self.emb_size
        return next(self.parameters()).new(1, self.batch_size, nh).zero_()

    def reset(self):
        "Reset the hidden states."
        self.hidden = [(self._one_hidden(l), self._one_hidden(l))
                       for l in range(self.n_layers)]

class LinearDecoder(nn.Module):
    """
    We add a top layer to the AWD LSTM. This is a linear model with dropout.
    """
    def __init__(self, n_out, n_hid, output_prob, tie_encoder=None, bias=True):
        super().__init__()
        self.output_dropout = RNNDropout(output_prob)
        self.decoder = nn.Linear(n_hid, n_out, bias=bias)
        if bias:
            self.decoder.bias.data.zero_()
        if tie_encoder:
            self.decoder.weight = tie_encoder.weight
        else:
            init.kaiming_uniform_(self.decoder.weight)

    def forward(self, input):
        raw_outputs, outputs = input
        # we call dropout first
        output = self.output_dropout(outputs[-1]).contiguous()
        # we call the linear model
        decoded = self.decoder(output.view(output.size(0)*output.size(1),
                                           output.size(2)))
        return decoded, raw_outputs, outputs

class SequentialRNN(nn.Sequential):
    "A sequential module that passes the reset call to its children."
    def reset(self):
        for child in self.children():
            if hasattr(child, 'reset'): child.reset()

def get_language_model(vocab_size, emb_size, n_hid, n_layers, pad_token,
                       output_prob=0.4, hidden_prob=0.2, input_prob=0.6,
                       embed_prob=0.1, weight_prob=0.5, tie_weights=True, bias=True):
    rnn_enc = AWD_LSTM(vocab_size, emb_size, n_hid=n_hid, n_layers=n_layers,
                       pad_token=pad_token, hidden_prob=hidden_prob,
                       input_prob=input_prob, embed_prob=embed_prob,
                       weight_prob=weight_prob)
    enc = rnn_enc.emb if tie_weights else None
    # the rnn_enc is the AWD LSTM
    # its output is passed to the top linear layer (with dropout)
    return SequentialRNN(rnn_enc,
                         LinearDecoder(vocab_size, emb_size, output_prob,
                                       tie_encoder=enc, bias=bias))

def lm_splitter(model):
    """
    Splits the language model provided by the get_language_model into multiple
    param groups to do transfer learning (e.g., from Wikipedia to IMDB):
    -- we have one group for each rnn + corresponding dropout, for a
    total of 2 if we had n_layers = 2 in the get_language_model call;
    -- we have one final group that contains the embeddings/decoder.
    The final group needs to be trained the most (new embedding vectors).
    """
    groups = []
    for i in range(len(model[0].rnns)):
        groups.append(nn.Sequential(model[0].rnns[i], model[0].hidden_dropouts[i]))
    groups += [nn.Sequential(model[0].emb, model[0].emb_dropout, model[0].input_dropout, model[1])]
    return [list(group.parameters()) for group in groups]

class AWD_LSTM1(nn.Module):
    """
    AWD-LSTM inspired by https://arxiv.org/abs/1708.02182,
    updated to deal with pad_packed_sequence and pack_padded_sequence.
    """
    initrange=0.1

    def __init__(self, vocab_size, emb_size, n_hid, n_layers, pad_token,
                 hidden_prob=0.2, input_prob=0.6, embed_prob=0.1, weight_prob=0.5):
        super().__init__()
        self.batch_size = 1
        self.emb_size = emb_size
        self.n_hid = n_hid
        self.n_layers = n_layers
        self.pad_token = pad_token
        self.emb = nn.Embedding(vocab_size, emb_size, padding_idx=pad_token)
        self.emb_dropout = EmbeddingDropout(self.emb, embed_prob)
        # we create n_layers of LSTMs
        self.rnns = [nn.LSTM(emb_size if l == 0 else n_hid,
                             (n_hid if l != n_layers - 1 else emb_size),
                             1, batch_first=True)
                     for l in range(n_layers)]
        self.rnns = nn.ModuleList([WeightDropout(rnn, weight_prob)
                                   for rnn in self.rnns])
        self.emb.weight.data.uniform_(-self.initrange, self.initrange)
        self.input_dropout = RNNDropout(input_prob)
        self.hidden_dropouts = nn.ModuleList([RNNDropout(hidden_prob)
                                              for l in range(n_layers)])

    def forward(self, input):
        batch_size, seq_len = input.size()
        if batch_size != self.batch_size:
            self.batch_size = batch_size
            self.reset()
        mask = (input == self.pad_token)
        lengths = seq_len - mask.long().sum(1)
        n_empty = (lengths == 0).sum()
        if n_empty > 0:
            input = input[:-n_empty]
            lengths = lengths[:-n_empty]
            self.hidden = [(h[0][:, :input.size(0)], h[1][:, :input.size(0)])
                           for h in self.hidden]
        raw_output = self.input_dropout(self.emb_dropout(input))
        new_hidden, raw_outputs, outputs = [], [], []
        for l, (rnn, hid_dropout) in enumerate(zip(self.rnns, self.hidden_dropouts)):
            # take data of different lengths and shape it to pass to RNN
            raw_output = pack_padded_sequence(raw_output, lengths, batch_first=True)
            raw_output, new_h = rnn(raw_output, self.hidden[l])
            # this is where the padding actually happens
            raw_output = pad_packed_sequence(raw_output, batch_first=True)[0]
            raw_outputs.append(raw_output)
            # we do hidden dropout for all layers but the last one
            if l != self.n_layers - 1: raw_output = hid_dropout(raw_output)
            outputs.append(raw_output)
            new_hidden.append(new_h)
        self.hidden = to_detach(new_hidden)
        return raw_outputs, outputs, mask

    def _one_hidden(self, l):
        "Return one hidden state."
        nh = self.n_hid if l != self.n_layers - 1 else self.emb_size
        return next(self.parameters()).new(1, self.batch_size, nh).zero_()

    def reset(self):
        "Reset the hidden states."
        self.hidden = [(self._one_hidden(l), self._one_hidden(l)) for l in range(self.n_layers)]

class Pooling(nn.Module):
    """
    This is just a pedagogically useful model. The actual pooling classifier is
    provided in PoolingLinearClassifier below.
    The LSTMs create hidden states for bptt time steps. We decide what to pass
    to the classifier here. Following concat pooling from vision, we use three
    things for the classification head of the model. We concatenate:
    -- the last hidden state
    -- the average (mean) pool of all the bptt hidden states
    -- the max pool of all the bptt hidden states
    We pass the resulting concatenated tensor to the classifier.
    """
    def forward(self, input):
        raw_outputs, outputs, mask = input
        # last hidden state
        output = outputs[-1]
        # once again, we need to ignore the padding in the last hidden state,
        # as well as the average pool and max pool tensors
        lengths = output.size(1) - mask.long().sum(dim=1)
        # average pool
        avg_pool = output.masked_fill(mask[:, :, None], 0).sum(dim=1)
        avg_pool.div_(lengths.type(avg_pool.dtype)[:, None])
        # max pool
        max_pool = output.masked_fill(mask[:, :, None], -float('inf')).max(dim=1)[0]
        # Concat pooling
        x = torch.cat([output[torch.arange(0, output.size(0)), lengths-1],
                       max_pool, avg_pool], 1)
        return output, x

def batchnorm_dropout_linear(n_in, n_out, batch_norm=True, dropout_prob=0.,
                             activation=None):
    "Creates a list of layers: batchnorm, dropout, linear, activation."
    layers = [nn.BatchNorm1d(n_in)] if batch_norm else []
    if dropout_prob != 0:
        layers.append(nn.Dropout(dropout_prob))
    layers.append(nn.Linear(n_in, n_out))
    if activation is not None:
        layers.append(activation)
    return layers

class PoolingLinearClassifier(nn.Module):
    """
    Create a linear classifier with pooling:
    -- the concat pooling layer, followed by
    -- a list of batchnorm + dropout + linear + activation layers
    """
    def __init__(self, layers, dropout_probs):
        super().__init__()
        modified_layers = []
        activations = [nn.ReLU(inplace=True)] * (len(layers) - 2) + [None]
        # list of batchnorm + dropout + linear layers
        for n_in, n_out, dropout_prob, activation in zip(layers[:-1], layers[1:],
                                                         dropout_probs, activations):
            modified_layers += batchnorm_dropout_linear(n_in, n_out,
                                                        dropout_prob=dropout_prob,
                                                        activation=activation)
        self.layers = nn.Sequential(*modified_layers)

    def forward(self, input):
        """
        The LSTMs create hidden states for bptt time steps. We decide what to pass
        to the classifier here. Following concat pooling from vision, we use three
        things for the classification head of the model. We concatenate:
        -- the last hidden state
        -- the average (mean) pool of all the bptt hidden states
        -- the max pool of all the bptt hidden states
        We pass the resulting concatenated tensor to the linear classifier.
        """
        raw_outputs, outputs, mask = input
        # last hidden state
        output = outputs[-1]
        # we need to ignore the padding in the last hidden state,
        # as well as the average pool and max pool tensors
        lengths = output.size(1) - mask.long().sum(dim=1)
        # average pool
        avg_pool = output.masked_fill(mask[:, :, None], 0).sum(dim=1)
        avg_pool.div_(lengths.type(avg_pool.dtype)[:, None])
        # max pool
        max_pool = output.masked_fill(mask[:, :, None], -float('inf')).max(dim=1)[0]
        # Concat pooling
        x = torch.cat([output[torch.arange(0, output.size(0)), lengths-1],
                       max_pool, avg_pool], 1)
        # pass the concat-pooled tensor through the linear layers
        x = self.layers(x)
        return x

def pad_tensor(t, batch_size, val=0.):
    if t.size(0) < batch_size:
        return torch.cat([t, val + t.new_zeros(batch_size-t.size(0), *t.shape[1:])])
    return t

class SentenceEncoder(nn.Module):
    "The encoder is the AWD LSTM model that gets called on the input text."
    def __init__(self, encoder, bptt, pad_idx=1):
        super().__init__()
        self.bptt = bptt
        self.encoder = encoder
        self.pad_idx = pad_idx

    def concat(self, arrs, batch_size):
        return [torch.cat([pad_tensor(l[si],batch_size) for l in arrs], dim=1)
                for si in range(len(arrs[0]))]

    def forward(self, input):
        batch_size, seq_len = input.size()
        self.encoder.batch_size = batch_size
        self.encoder.reset()
        raw_outputs, outputs, masks = [], [], []
        # We go through the input one bptt at a time
        for i in range(0, seq_len, self.bptt):
            # we call the RNN model on it
            r, o, m = self.encoder(input[:,i: min(i+self.bptt, seq_len)])
            # we keep appending the results
            masks.append(pad_tensor(m, batch_size, 1))
            raw_outputs.append(r)
            outputs.append(o)
        return self.concat(raw_outputs, batch_size), self.concat(outputs, batch_size),torch.cat(masks,dim=1)

def get_text_classifier(vocab_size, emb_size, n_hid, n_layers, n_out, pad_token,
                        bptt, output_prob=0.4, hidden_prob=0.2, input_prob=0.6,
                        embed_prob=0.1, weight_prob=0.5, layers=None, dropout_probs=None):
    "To create a full AWD-LSTM"
    rnn_enc = AWD_LSTM1(vocab_size, emb_size, n_hid=n_hid, n_layers=n_layers,
                        pad_token=pad_token, hidden_prob=hidden_prob,
                        input_prob=input_prob, embed_prob=embed_prob,
                        weight_prob=weight_prob)
    enc = SentenceEncoder(rnn_enc, bptt)
    if layers is None:
        layers = [50]
    if dropout_probs is None:
        dropout_probs = [0.1] * len(layers)
    layers = [3 * emb_size] + layers + [n_out]
    dropout_probs = [output_prob] + dropout_probs
    return SequentialRNN(enc, PoolingLinearClassifier(layers, dropout_probs))

def class_splitter(model):
    enc = model[0].encoder
    groups = [nn.Sequential(enc.emb, enc.emb_dropout, enc.input_dropout)]
    for i in range(len(enc.rnns)):
        groups.append(nn.Sequential(enc.rnns[i], enc.hidden_dropouts[i]))
    groups.append(model[1])
    return [list(group.parameters()) for group in groups]



from .utils import *
from .datasets import *
from .data_block import *
from .metrics import *
from .callbacks import *
from .hooks import *
from .optimizer import *
from .awd_lstm import *
from .learner import *
from .rlearner import *



import torch
from functools import partial
from .utils import listify, compose

def maybe_update(sources, destination, func):
    """
    Updates defaults (values) in destination with values from func
    applied to sources, only if defaults (values) not already specified.
    """
    for source in sources:
        for key, value in func(source).items():
            if key not in destination:
                destination[key] = value

def get_defaults(default):
    return getattr(default, '_defaults', {})

class Optimizer():
    """
    param_groups is assumed to be a list of lists for discriminative learning rates.
    hypers are hyperparams for each param group (each list in the list of lists).
    """
    def __init__(self, param_groups, steppers, **defaults):
        self.steppers = listify(steppers)
        maybe_update(self.steppers, defaults, get_defaults)
        # might be a generator
        self.param_groups = list(param_groups)
        # ensure param_groups is a list of lists
        if not isinstance(self.param_groups[0], list):
            self.param_groups = [self.param_groups]
        # list of dicts of hyperparams, one dict per param group;
        # each dict clones the provided **defaults
        self.hypers = [{**defaults} for param_group in self.param_groups]

    def grad_params(self):
        "Return list of (param, hyper) tuples if param has gradient."
        param_hyper_list = []
        for param_group, hyper in zip(self.param_groups, self.hypers):
            for param in param_group:
                if param.grad is not None:
                    param_hyper_list.append((param, hyper))
        return param_hyper_list

    def zero_grad(self):
        for param, hyper in self.grad_params():
            param.grad.detach_()
            param.grad.zero_()

    def step(self):
        for param, hyper in self.grad_params():
            compose(param, self.steppers, **hyper)

def sgd_step(param, lr, **kwargs):
    "Example stepper for use with Optimizer."
    param.data.add_(-lr, param.grad.data)
    return param
# sgd_step._defaults = dict(lr=1e-1)

def weight_decay(param, lr, weight_decay_param, **kwargs):
    "Update with gradient of L2 regularizer with weight_decay_param."
    param.data.mul_(1 - lr * weight_decay_param)
    return param
weight_decay._defaults = dict(weight_decay_param=0.)

# alternatively
# def l2_reg(param, lr, weight_decay_param, **kwargs):
    # param.grad.data.add_(weight_decay_param, param.data)
    # return param
# l2_reg._defaults = dict(weight_decay_param=0.)

# SGD optimizer with weight decay
sgd_opt = partial(Optimizer, steppers=[weight_decay, sgd_step])

# To inspect the hyperparams of the optimizer:
# opt = sgd_opt(learn.model.parameters(), lr=1e-1, weight_decay_param=1e-4)
# print(opt.hypers)

class StatefulOptimizer(Optimizer):
    """
    Optimizer with state for momentum.
    We assemble the relevant state with stats.
    """
    def __init__(self, param_groups, steppers, stats=None, **defaults):
        self.stats = listify(stats)
        maybe_update(self.stats, defaults, get_defaults)
        super().__init__(param_groups, steppers, **defaults)
        self.state = {}

    def step(self):
        for param, hyper in self.grad_params():
            if param not in self.state:
                #Create a state for param and call all the statistics to initialize it.
                self.state[param] = {}
                maybe_update(self.stats, self.state[param],
                             lambda grad_stat: grad_stat.init_state(param))
            state = self.state[param]
            for stat in self.stats:
                state = stat.update(param, state, **hyper)
            compose(param, self.steppers, **state, **hyper)
            self.state[param] = state

class Stat():
    "Base class used to assemble state for StatefulOptimizer."
    _defaults = {}
    def init_state(self, param):
        raise NotImplementedError
    def update(self, param, state, **kwargs):
        raise NotImplementedError

class AverageGrad(Stat):
    "Assembles state for momentum."
    _defaults = dict(mom=0.9)
    def __init__(self, dampening=True):
        self.dampening = dampening

    def init_state(self, param):
        return {'grad_avg': torch.zeros_like(param.grad.data)}

    def update(self, param, state, mom, **kwargs):
        state['mom_damp'] = 1 - mom if self.dampening else 1.
        # the update below is just the definition of momentum
        state['grad_avg'].mul_(mom).add_(state['mom_damp'],
                                         param.grad.data)
        return state

def momentum_step(param, lr, grad_avg, **kwargs):
    param.data.add_(-lr, grad_avg)
    return param

sgd_mom_opt = partial(StatefulOptimizer,
                      steppers=[momentum_step, weight_decay],
                      stats=AverageGrad(dampening=False),
                      weight_decay_param=1e-2)

class AverageSqrGrad(Stat):
    _defaults = dict(sqr_mom=0.99)
    def __init__(self, dampening=True):
        self.dampening = dampening

    def init_state(self, param):
        return {'sqr_avg': torch.zeros_like(param.grad.data)}

    def update(self, param, state, sqr_mom, **kwargs):
        state['sqr_damp'] = 1 - sqr_mom if self.dampening else 1.
        state['sqr_avg'].mul_(sqr_mom).addcmul_(state['sqr_damp'],
                                                param.grad.data,
                                                param.grad.data)
        return state

class StepCount(Stat):
    def init_state(self, param):
        return {'step': 0}

    def update(self, param, state, **kwargs):
        state['step'] += 1
        return state

def debias(mom, damp, step):
    return damp * (1 - mom**step) / (1 - mom)

def adam_step(param, lr, mom, mom_damp, step, sqr_mom, sqr_damp,
              grad_avg, sqr_avg, eps, **kwargs):
    debias_mom = debias(mom, mom_damp, step)
    debias_sqr_mom = debias(sqr_mom, sqr_damp, step)
    param.data.addcdiv_(-lr / debias_mom, grad_avg,
                        (sqr_avg/debias_sqr_mom).sqrt() + eps)
    return param
adam_step._defaults = dict(eps=1e-5)

def adam_opt(extra_step=None, **kwargs):
    return partial(StatefulOptimizer,
                   steppers = [adam_step, weight_decay] + listify(extra_step),
                   stats = [AverageGrad(), AverageSqrGrad(), StepCount()],
                   **kwargs)



from setuptools import setup

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name='lgm',
    version='0.0.10',
    description='Lightweight python3 library for natural LanGuage Modeling (LGM). Mostly useful for DL/RL experiments with natural language data. [originally based on the fastai course v3, dl2 (March-April 2019)]',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/abrsvn/lgm',
    author='Adrian Brasoveanu',
    author_email='abrsvn@gmail.com',
    license='Apache License Version 2.0',
    packages=['lgm'],
    zip_safe=False
)



import torch
from functools import partial
from torch import tensor
from .optimizer import adam_opt
from .utils import ALL_CALLBACKS, param_getter, listify, CancelTrainException, CancelEpochException, CancelBatchException
from .callbacks import TrainEvalCallback, ProgressCallback, Plotter, SendToDeviceCallback, AvgStatsCallback, SaveModelCallback

class RLearner():
    """
    THIS IS JUST A STUB (for now, mostly copied from learner.py) FOR A REINFORCEMENT LEARNER CLASS
    -- we replace model with agent and data with env (environment)
    -- restructure this in a way partly inspired by rl_glue code from the Coursera RL specialization
    """
    def __init__(self, agent, env, loss_func, opt_func=adam_opt(), lr=1e-2,
                 splitter=param_getter, metrics=None, callbacks=None,
                 callback_funcs=None, device=None, reset_opt=False,
                 model_name=None, path_str=None):
        raise NotImplementedError
